// +--------------------------------------------------------------+
// |                                                              |
// |  SNA2QUIL V1.0 by Zeljko Juric                               |
// |  Heavily based on UNQUILL V0.7.5 by John Elliot              |
// |                                                              |
// |  Converts ZX Spectrum .SNA snapshots to the format which is  |
// |  usable with The Quill Adventure Writer for TI-89 & TI-92+   |
// |                                                              |
// |  Tested with Borland TURBO C; probably works with any other  |
// |  ANSI C compiler                                             |
// +--------------------------------------------------------------+

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef unsigned char byte;
typedef unsigned short word;

byte zmem(word);
word zword(word);
void listitems(char*,word,word);
void listwords(word);
void listcond(word,word);
void listconn(word,byte);
void listobj(word,word,word,word);
void listgfx(void);
void opword(word);
void opcond(word*);
void opact(word*);
void opch32(char);
void opcact(char,char*);
void expch(byte,word*);
void expdict(byte,word*);
void sysmess(byte);

// In all games created with The Quill for ZX Spectrum, somewhere in memory
// (0x6D04 for vanilla Quill with no Illustrator/Patch/Press extensions) there
// is a color definition table which looks like
//
// BYTE 0x10,ink,0x11,paper,0x12,flash,0x13,bright,0x14,inverse,0x15,over
// BYTE border
//
// SNA2QUIL use this as a magic number to find the Quill database in memory.
// After the colour table is a table of values:
//
// BYTE number of objects conveyable
// BYTE number of objects
// BYTE number of locations
// BYTE number of messages
//
// then a table of pointers:
//
// WORD rsptab    ; Response table (Actions)
// WORD protab    ; Process table (Events)
// WORD objtop    ; Top of object descriptions
// WORD loctop    ; Top of locations.
// WORD msgtop    ; Top of messages.
// **             ; (see below)
// WORD conntab   ; Table of connections.
// WORD voctab    ; Table of words.
// WORD oloctab   ; Table of object inital locations.
// WORD free      ; [Early games] Start of free memory
//                ; [Late games]  Word/object mapping
// WORD freeend   ; End of free memory
// BYTE ?,?,?     ; ?
// BYTE 0x80,dict ; Expansion dictionary (for compressed games).
//
// In later games, there is an extra byte (the number of system messages)
// after the number of messages, and at the label ** the following word is
// inserted:
//
// WORD	systab	  ; Points to word pointing to table of system messages.
//
// The address of the user-defined graphic characters is stored at 23675. In
// "Early" games, the system messages are at UDGs+168.

byte zxmemory[0xA400];   // Spectrum memory from 0x5C00 up
char xpos=0;             // Used in msg display
char islql=0;            // 1 for later data records, 0 for earlier ones
char lstgfx=0;           // 1 if converting pictures is requested
word ofarg=0;            // Output to file or stdout?
word vocab,dict;         // Spectrum address of vocabulary & dictionary
word loctab;             // Tables of locations
word objtab;             // Tables of objects
word msgtab;             // Tables of messages
word sysbase;            // Base of system messages
word conntab;            // Connections table
word objmap;             // Word-to-object map
char *inname;            // Input filename

byte maxcar;             // Max number of objects conveyable
byte nobj;               // Number of objects
byte nsys;               // Number of system messages

FILE *infile,*outfile;

#define AV10 argv[1][0]
#define AV11 argv[1][1]

void main(int argc,char *argv[])
{
  word n,zxptr;
  byte found=0;
  byte nmsg,nloc;
  char *ofptr;
  if(argc==1||((AV10=='-'||AV10=='/')&&   // If the command looks like a help,
    (AV11=='h'||AV11=='H'||AV11=='?')))   // then print the helpscreen
      {
        fprintf(stderr,"\nSNA2QUIL V1.0 by Zeljko Juric.\n"
          "Heavily based on UNQUILL V0.7.5 by John Elliot.\n\n"
          "Converts .SNA snapshots of Quilled games to the format usable with"
          "\nThe Quill Adventure Writer for TI-89 and TI-92+\n\n"
          "Command format is SNA2QUIL input-file [-Ooutput-file] "
          "[-opt -opt ...]"
          "\n\nOptions are:\n"
          "-L  : Attempt to interpret as a 'later' type Quill file.\n"
          "-G  : Extract also pictures data, if possible.\n"
          "-GT : Extract pictures data only.\n"
          "-O  : Redirect output to a file. If this option is not\n"
          "      present, stdout is used.\n\n"
          "Mail bug reports to zjuric@utic.net.ba\n");
        exit(1);
      }
  for(n=2;n<argc;n++)                         // Parse for options.
    {                                         // The option formats are:
      if(argv[n][0]=='-'||argv[n][0]=='/')    //  -option (Unix)
        {                                     //  /option (DOS)
          switch (argv[n][1])
            {
              case 'O':
                ofarg=n;
                ofptr=(argv[n]+2);
                break;
              case 'L':
                islql++;
                break;
              case 'G':
                lstgfx=1;
                if(argv[n][2]=='T') lstgfx=2;
                break;
              default:
                fprintf(stderr,"SNA2QUIL: Invalid option: %s\n",argv[n]);
                exit(1);
            }
        }
    }
  if(ofarg)                                 // If output file was specified,
    {                                       // then select it
      if((outfile=fopen(ofptr,"w"))==NULL)
        {
          perror(ofptr);
          exit(1);
        }
    }
  else outfile=stdout;
  inname=argv[1];                           // Load the snapshot. To save
  if((infile=fopen(inname,"rb"))==NULL)     // the space, the printer buffer
    {                                       // and the screen memory will be
      perror(inname);                       // ignored, because they can not
      exit(1);                              // contain anything of value
    }
  if(fseek(infile,0x1C1B,SEEK_SET))         // Skip screen, printer buffer
    {                                       // and system variables, then
      perror(inname);                       // load the rest
      exit(1);
    }
  if(fread(zxmemory,sizeof(zxmemory),1,infile)!=1)     // Load the file
    {
      perror(inname);
      exit(1);
    }
  for(n=0x5C00;n<0xFFF5;n++)    // .SNA read ok. Find a Quill signature
    {
      if(zmem(n)==16&&zmem(n+2)==17&&zmem(n+4)==18&&zmem(n+6)==19&&
        zmem(n+8)==20&&zmem(n+10)==21)
          {
            fprintf(stderr,"Quill signature found.\n");
            found=1;
            break;
          }
    }
  if(!found)
    {
      fprintf(stderr,"%s does not seem to be a valid Quill .SNA file.\n"
        "If you know it is, you have found a bug in SNA2QUIL.\n",inname);
      exit(1);
    }
  zxptr=n+13;
  maxcar=zmem(zxptr);
  nobj=zmem(zxptr+1);
  nloc=zmem(zxptr+2);
  nmsg=zmem(zxptr+3);
  if(islql)
    {
      zxptr+=1;
      nsys=zmem(zxptr+3);
      vocab=zword(zxptr+18);
      dict=zxptr+29;
    }
  else vocab=zword(zxptr+16);
  objtab=zword(zxptr+8);
  loctab=zword(zxptr+10);
  msgtab=zword(zxptr+12);
  if(islql) sysbase=zword(zxptr+14);
  else sysbase=zword(23675)+168;
  conntab=zword(zxptr+14+2*(islql));
  if(islql) objmap=zword(zxptr+22);
  if(lstgfx==2)
    {
      fprintf(outfile,"$$TITLE(\"PIC:%s\")\n\n",inname);
      fprintf(outfile,"$$PICTURES_TEST(%d)\n\n",nloc);
    }
  else
    {
      fprintf(outfile,"$$TITLE(\"%s\")\n\n",inname);
      listwords(vocab);
      fputs("$$LOCATIONS\n",outfile);
      listitems("LOC",zword(loctab),nloc);
      fputs("\n$$MESSAGES\n",outfile);
      listitems("MSG",zword(msgtab),nmsg);
      if(islql) fputs("\n$$SYSTEM_MESSAGES\n",outfile);
      else fputs("\n$$OLDSTYLE_SYSTEM_MESSAGES\n",outfile);
      if(islql) listitems("MSG",zword(zword(zxptr+14)),nsys);
      else listitems("MSG",zword(23675)+168,32);
      fprintf(outfile,"\n$$OBJECTS(%d)\n",maxcar);
      listobj(zword(objtab),objmap,zword(zxptr+18+2*islql),nobj);
      listconn(zword(conntab),nloc);
      fputs("\n$$EVENTS\n\nEVENT\n    ISDESC\n    CLS\n    SHOWLOC\n"
        "    NEWLINE\n    LISTOBJ\n",outfile);
      listcond(zword(zxptr+6),1);
      fputs("\n$$ACTIONS\n\nACTION(_,_)\n    TRYMOVE\n    DESC\n",outfile);
      listcond(zword(zxptr+4),0);
      fputs("\n$$PICTURES\n\n",outfile);
    }
  if(lstgfx) listgfx();
  if(lstgfx!=2) fputs("$$EXTERN\n\n$$END\n",outfile);
  else fputs("$$END_TEST\n",outfile);
  if(fclose(infile))
    {
    perror(inname);
    exit(1);
    }
  if(ofarg)
    if(fclose(outfile))
      {
        perror(ofptr);
        exit(1);
      }
}

// All Spectrum memory accesses are routed through this procedure.

byte zmem(word addr)
{
  if(addr<0x5C00)
    {
      fprintf(stderr,"\nInvalid address %4.4x requested. Try %s the -L "
        "option.",addr,islql?"omitting":"including");
      exit(1);
    }
  return zxmemory[addr-0x5C00];
}

word zword(word addr)
{
  return zmem(addr)+256*zmem(addr+1);
}

void opch32(char ch)
{
  char x;
  if(ch=='\"') fputs("\\\"",outfile);
  else if(ch=='\\') fputs("\\\\",outfile);
  else if(ch!='\n') fputc(ch,outfile);
  if(ch=='\n')
    {
      xpos=0;
      fputs("\\n\"\n    \"",outfile);
    }
  else if(ch==8)
    {
      xpos--;
      fseek(outfile,-1,SEEK_CUR);
    }
  else if(xpos>=31)
    {
      xpos=0;
      fprintf(outfile,"\"\n    \"");
    }
  else xpos++;
}

// Response and process table have the following format:
//
// rsptab: ...   ; response  - both stored as: BYTE verb,noun
// protab: ...   ; process                     WORD address-of-handler
//
// The table is terminated by verb 0. The handler format is:
//
// DB condition1,condition2,...,0xFF
// DB action1,action2,...,0xFF

void listcond(word first,word status)
{
  word addr=first,subsaddr;
  word lastw1=256,lastw2=256;
  while(zmem(addr))
    {
      if(!status)
        {
          byte w1,w2;
          w1=zmem(addr++); w2=zmem(addr++);
          if(w1==lastw1&&w2==lastw2) fputs("  ELSE\n",outfile);
          else
            {
              fputs("ACTION(",outfile);
              opword(w1);
              fputc(',',outfile);
              opword(w2);
              fprintf(outfile,")\n");
            }
          lastw1=w1; lastw2=w2;
        }
      else
        {
          fputs("EVENT\n",outfile);
          addr+=2;
        }
      subsaddr=zword(addr++);
      if(zmem(subsaddr)!=255)
        {
          while(zmem(subsaddr)!=255) opcond(&subsaddr);
        }
      subsaddr++;
      while(zmem(subsaddr)!=255) opact(&subsaddr);
      addr++;
    }
}

void opcond(word *condad)
{
  byte id=zmem((*condad)++);
  byte param1;
  opcact(id,"AT     NOTAT  ATGT   ATLT   PRESENTABSENT WORN   NOTWORNCARRIED"
    "NOTCARRCHANCE ZERO   NOTZEROEQ     GT     LT     ");
  param1=zmem((*condad)++);
  fprintf(outfile,"(%d",param1);
  if(id>12) fprintf(outfile,",%d",zmem((*condad)++));
  fputs(")\n",outfile);
}

void opact(word *actad)
{
  static char inited=0;
  static char params[40];
  char n;
  byte id=zmem((*actad)++);
  byte actpar[2];
  if(!inited&&islql)
    {
      for(n=0;n<39;n++)
        {
          if(n<17) params[n]=0;
          else if(n<33) params[n]=1;
          else params[n]=2;
        }
      params[29]=2;
      params[30]=2;
      inited++;
    }
  if(!inited&&!islql)
    {
      for(n=0;n<39;n++)
        {
          if(n<11) params[n]=0;
          else if(n<23) params[n]=1;
          else params[n]=2;
        }
      params[20]=2;
      inited++;
    }
  if(islql) opcact(id,"INVEN  DESC   QUIT   END    DONE   OK     ANYKEY "
    "SAVE   LOAD   TURNS  SCORE  CLS    DROPALLAUTOG  AUTOD  AUTOW  AUTOR  "
    "PAUSE  PAPER  INK    BORDER GOTO   MESSAGEREMOVE GET    DROP   WEAR   "
    "DESTROYCREATE SWAP   PLACE  SET    CLEAR  PLUS   MINUS  LET    BEEP   ");
  else opcact(id,"INVEN  DESC   QUIT   END    DONE   OK     ANYKEY SAVE   "
    "LOAD   TURNS  SCORE  PAUSE  GOTO   MESSAGEREMOVE GET    DROP   WEAR   "
    "DESTROYCREATE SWAP   "
    "SET    CLEAR  PLUS   MINUS  LET    BEEP   ");
  n=params[id];
  if(n)
    {
      fputc('(',outfile);
      while(n>0)
        {
          actpar[n-1]=(zmem((*actad)++));
          fprintf(outfile,"%d",actpar[--n]);
          if(n) fputc(',',outfile);
        }
      fputc(')',outfile);
    }
  fputc('\n',outfile);
}

void opcact(char n,char *str)
{
  word i=0;
  str+=7*n;
  fputs("    ",outfile);
  while(*str&&*str!=' '&&i<7)
    {
      fputc(*str++,outfile);
      i++;
    }
}

// Connections are stored as follows:
//
// conntab: DB word,room,word,room,...,0xFF
//          DB word,room,word,room,...,0xFF
//          ...

void listconn(word first,byte nrms)
{
  byte item=0;
  word n=0;
  fputs("\n$$CONNECTIONS\n\n",outfile);
  xpos=0;
  n=first;
  for(item=0;item<nrms;item++)
    {
      if(zmem(n)!=255) fprintf(outfile,"FROM(%d)\n",item);
      while(zmem(n)!=255)
        {
          fputs("    CONN(",outfile);
          opword(zmem(n++));
          fprintf(outfile,",%d)\n",zmem(n++));
        }
      n++;
    }
}

// The vocabulary has the following format:
//
// voctab: BYTE 'word',number
//         BYTE 'word',number
//         ...
//
// Each word is exactly 4 bytes long. Words are complemented to deter hackers.
// Table terminated with a word entry all five bytes of which are 0 before
// decoding.

void listwords(word first)
{
  word n,m;
  fputs("$$VOCABULARY(4)\n\n",outfile);
  for(n=first;zmem(n);n=n+5)
    if(zmem(n+4)!=255)
      {
        fprintf(outfile,"WORD(%d,",zmem(n+4));
        for(m=0;m<4&&zmem(m+n)!=223;m++) fputc(255-zmem(m+n),outfile);
        fputs(")\n",outfile);
      }
  fputc('\n',outfile);
}

void opword(word wrd)
{
  word n,m;
  for(n=vocab;zmem(n);n+=5)
    {
      if(zmem(n+4)==wrd)
       {
         for(m=0;m<4;m++)
           {
             word ch=255-zmem(m+n);
             if(ch!=' ') fputc(ch,outfile);
           }
         break;
       }
    }
}

// Location texts, messages and system messages are organized as follows:
//
// loctab: ...		;locations
// loctop: WORD	loctab
//	   WORD	locno1
//	   WORD	locno2
//         ...
//msgtab:  ...		;messages
//msgtop:  WORD	msgtab
//	   WORD	mess1
//	   WORD	mess2
//         ...
//
// All object texts are stored with all bytes complemented to deter causal
// hackers. Texts are terminated with 0xE0 (i.e. 0x1F after decoding).

void listitems(char *itmid,word first,word howmany)
{
  byte item=0,cch;
  word n;
  howmany=howmany&255;
  fprintf(outfile,"\n%s(0,\"",itmid);
  xpos=0;
  n=first;
  for(item=0;item<howmany;)
    {
      cch=(255-(zmem(n++)));
      if(cch==31)                  // End of text
        {
          if(++item==howmany) fprintf(outfile,"\")\n");
          else fprintf(outfile,"\")\n%s(%d,\"",itmid,item);
          xpos=0;
        }
      else expch(cch,&n);
    }
}

// Objects are stored as follows:
//
// objtab:  ...	         ; object texts
// objtop:  WORD objtab
//	    WORD object1
//	    WORD object2
//          ...
// oloctab: ...		 ;initial locations of objects.
//
// All object texts are stored with all bytes complemented to deter causal
// hackers. Texts are terminated with 0xE0 (i.e. 0x1F after decoding).
// Initial locations of objects may contain following special values:
// 0xFC => not created, 0xFD => worn, 0xFEh=> carried, 0xFF => end of table

void listobj(word first,word fmap,word fpos,word howmany)
{
  byte item=0,cch,wrd,loc;
  word n;
  howmany=howmany&255;
  fputs("\nOBJ(0,\"",outfile);
  xpos=0;
  n=first;
  for(item=0;item<howmany;)
    {
      cch=(255-(zmem(n++)));
      if(islql) wrd=zmem(fmap+item);
      loc=zmem(fpos+item);
      if(cch==31)                  // End of text
        {
          fputs("\",",outfile);
          if(islql) opword(wrd);
          else fputc('_',outfile);
          fputc(',',outfile);
          if(loc==254) fputs("$CARRIED",outfile);
          else if(loc==253) fputs("$WORN",outfile);
          else if(loc==252) fputs("$NOWHERE",outfile);
          else fprintf(outfile,"%d",loc);
          fputs(")\n",outfile);
          if(++item!=howmany) fprintf (outfile,"OBJ(%d,\"",item);
          xpos=0;
        }
      else expch(cch,&n);
    }
}

// Dictionary is stored in ASCII, high bit set on last letter of each word.

void expdict(byte cch)
{
  word d=dict,dummy;
  if(islql)        // Early games aren't compressed & have no dictionary
    {
      cch-=164;
      while(cch)
        if(zmem(d++)>0x7f) cch--;                         // d is the address
      while(zmem(d)<0x80) expch(zmem(d++)&0x7F,&dummy);   // of the expansion
      expch(zmem(d)&0x7F,&dummy);                         // text
    }
}

void expch(byte cch,word *n)
{
  word tbsp;
  if(cch>31&&cch<127) opch32(cch);
  else if(cch>164) expdict(cch);
  else if(cch>126) opch32('?');
  else if(cch==6)
    for(opch32(' ');xpos%16;opch32(' '));
  else if(cch==8) opch32(8);
  else if(cch==13) opch32('\n');
  else if(cch==23)
    {
      tbsp=(255-zmem((*n)++))&31;
      ++(*n);
      if(xpos>tbsp) opch32('\n');
      for(;tbsp>0;tbsp--) opch32(' ');
    }
  else if(cch>15&&cch<22) (*n)++;
}

// 2 is a lousy sample size, but it seems that all the Spectrum Quill games
// illustrated with The Illustrator have their graphics tables at 0xFAB8.

char *pmov[]={"UP","UP_RIGHT","RIGHT","DOWN_RIGHT","DOWN","DOWN_LEFT","LEFT",
  "UP_LEFT"};

void listgfx(void)
{
  word n,m;
  word gfx_base=0xFAB8;	                 // Arbitrary???
  word gfx_ptrs=zword(gfx_base);
  word gfx_flags=zword(gfx_base+2);
  word gfx_count=zmem (gfx_base+6);
  byte gflag,nargs,value;
  word gptr;
  char *inv,*ovr;
  char *opcode;
  word neg[8];
  char tempop[30];
  for(n=0;n<gfx_count;n++)
    {
      gflag=zmem(gfx_flags+n);
      gptr=zword(gfx_ptrs+2*n);
      if(zmem(gptr)!=7)
        {
          fprintf(outfile,"DRAWING(%d,%s)\n",n,
            gflag&0x80?(lstgfx==2?"$NOWHERE":"$FULLSCR"):"$SUBROUTINE");
          do
    	    {
	      memset(neg,0,sizeof(neg));
	      gflag=zmem(gptr);
	      inv=""; ovr="";
	      if(gflag&0x08) ovr="XOR_";
	      if(gflag&0x10) inv="INV_";
	      value=(gflag>>3);
              switch(gflag&7)
	        {
	          case 0:
                    nargs=2;
                    opcode=tempop;
                    sprintf(opcode,"    %s%sPLOT(",ovr,inv);
                    if(*ovr=='X'&&*inv=='I') strcpy(opcode,"    AMOVE(");
                    break;
    	          case 1:
                    nargs=2;
                    if(gflag&0x40) neg[0]=1;
                    if(gflag&0x80) neg[1]=1;
                    if(*ovr=='X'&&*inv=='I') opcode="    MOVE(";
		    else
		      {
		        opcode=tempop;
           	        sprintf(opcode,"    %s%sLINE(",ovr,inv);
		      }
		    break;
	          case 2:
                    if(gflag&0x20)
		      {
                        if(gflag&0x40) neg[0]=1;
                        if(gflag&0x80) neg[1]=1;
		        nargs=3; opcode="    SHADE(";
		      }
                    else if(gflag&0x10)
		      {
                        nargs=4; opcode="    BLOCK(";
                      }
		    else
		      {
                        if(gflag&0x40) neg[0]=1;
                        if(gflag&0x80) neg[1]=1;
		        nargs=2; opcode="    FILL(";
		      }
		    break;
	          case 3:
                    nargs=1;
		    sprintf(tempop,"    CALL(%d,",value&7);
		    opcode=tempop;
		    break;
    	          case 4:
                    nargs=0;
		    sprintf(tempop,"    %s%sRPLOT(%s)",ovr,inv,pmov[value/4]);
		    opcode=tempop;
		    break;
	          case 5:
                    nargs=0;
		    opcode=tempop;
                    if(gflag&0x80) sprintf(tempop,"    BRIGHT(%d)",value&15);
		    else sprintf(tempop,"    PAPER(%d)",value&15);
		    break;
                  case 6:
                    nargs=0;
		    opcode=tempop;
                    if(gflag&0x80) sprintf(tempop,"    FLASH(%d)",value&15);
                    else sprintf(tempop,"    INK(%d)",value&15);
		    break;
	          case 7:
                    nargs=0; opcode="END_DRAWING\n";
                    break;
	        }
	      fprintf(outfile,"%s",opcode);
	      for(m=0;m<nargs;m++)
	        {
                  word arg=zmem(gptr+1+m);
                  if(opcode[4]!='B'&&m==1)
                    {
                      if(!(gflag&7)) arg=175-arg;
                      else neg[1]=1-neg[1];
                      if(!arg) neg[1]=0;
                    }
                  if(opcode[4]=='S'&&m==2)
                    {
                      byte i,bits[8];
                      for(i=0;i<8;i++)
                        {
                          bits[i]=arg&1; arg>>=1;
                        }
                      arg=257*(bits[7]+(bits[3]<<1)+(bits[5]<<2)+(bits[1]<<3)+
                        (bits[6]<<4)+(bits[2]<<5)+(bits[4]<<6)+(bits[0]<<7));
                      fprintf(outfile,"0x%04X",arg);
                    }
                  else fprintf(outfile,"%s%d",neg[m]?"-":"",arg);
                  fputc(m==nargs-1?')':',',outfile);
	        }
	      fputc('\n',outfile);
              gptr+=nargs+1;
            }
          while((gflag&7)!=7);
        }
      else gptr++;
    }
}
